/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComScale.cpp
    \brief    Functions for upscale and downscale of pictures for dynamic resolution coding
*/

#include <TComScale.h>
#include <assert.h>

static int fv_lut[512]=
{
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,
    7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7,

    6,6,6,6,6,6,6,6, 6,6,6,6,6,6,6,6, 6,6,6,6,6,6,6,6, 6,6,6,6,6,6,6,6,

    5,5,5,5,5,5,5,5, 5,5,5,5,5,5,5,5,

    4,4,4,4,4,4,4,4, 4,4,4,4,4,4,4,4,

    0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,

    1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,

    2,2,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,2,2,

    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3,
};

int getClassIndex( int d[], int size)
{
  int index = 0;

  for (int j=0; j<size; ++j){
    index += fv_lut[d[j]+255] << (3*j);
  }

  return index;
}

static int tapsD1[ 512 ][ 4 ] = { {-12, 77, 77, -13}, {-11, 64, 87, -13}, {-10, 47, 101, -12}, {-11, 35, 92, 9}, {-7, 77, 72, -15}, {0, 72, 87, -30}, {2, 60, 98, -34}, {-3, 39, 124, -34}, 
{-22, 93, 82, -23}, {-14, 71, 100, -29}, {-13, 62, 114, -34}, {-13, 53, 105, -18}, {-2, 76, 85, -30}, {-10, 69, 105, -36}, {4, 58, 106, -44}, {1, 41, 121, -41}, 
{-36, 111, 103, -46}, {-20, 85, 116, -49}, {-15, 70, 121, -46}, {-16, 67, 133, -55}, {-10, 90, 105, -51}, {-7, 84, 104, -47}, {-21, 71, 132, -53}, {8, 53, 124, -67}, 
{-50, 125, 132, -73}, {-31, 99, 136, -71}, {-22, 84, 143, -72}, {-16, 76, 148, -72}, {-28, 109, 156, -99}, {-10, 99, 132, -80}, {-8, 105, 127, -78}, {-2, 92, 128, -78}, 
{-3, 66, 69, -5}, {-5, 70, 63, 0}, {-7, 62, 56, 15}, {-7, 38, 56, 38}, {-15, 73, 77, -7}, {-6, 54, 95, -15}, {-3, 38, 110, -19}, {-4, 37, 117, -24}, 
{6, 58, 67, -1}, {-5, 66, 72, -4}, {-7, 77, 56, 0}, {-3, 89, 19, 18}, {-31, 88, 73, 0}, {-15, 66, 85, -6}, {-8, 50, 102, -17}, {-5, 44, 107, -21}, 
{12, 41, 91, -7}, {6, 51, 81, -4}, {-13, 65, 86, -8}, {-5, 79, 58, -7}, {-50, 105, 89, -10}, {-29, 80, 82, -4}, {-13, 60, 98, -19}, {-8, 46, 107, -20}, 
{21, 45, 148, -70}, {134, -46, 114, -50}, {35, 34, 107, -31}, {-1, 68, 128, -63}, {-100, 159, 102, -25}, {-35, 91, 106, -33}, {-26, 77, 100, -26}, {-12, 61, 122, -55}, 
{-16, 93, 58, -7}, {-10, 74, 71, -8}, {-17, 78, 73, -8}, {-17, 56, 73, 11}, {-14, 93, 52, -4}, {-9, 71, 65, -3}, {7, 56, 78, -21}, {-5, 49, 74, 1}, 
{-29, 103, 71, -17}, {-20, 84, 85, -21}, {-14, 75, 79, -12}, {-15, 60, 110, -27}, {-9, 87, 66, -15}, {-4, 74, 68, -13}, {-11, 60, 80, -9}, {3, 51, 76, -15}, 
{-41, 113, 86, -29}, {-29, 98, 91, -30}, {-17, 73, 99, -27}, {-30, 89, 122, -54}, {-12, 91, 76, -25}, {6, 74, 86, -35}, {-15, 73, 90, -22}, {4, 49, 98, -36}, 
{-67, 141, 92, -37}, {-46, 115, 95, -36}, {-35, 95, 119, -51}, {-29, 88, 95, -27}, {-43, 120, 83, -31}, {15, 69, 89, -43}, {10, 88, 94, -52}, {-2, 90, 73, -29}, 
{-1, 66, 67, -4}, {-12, 72, 72, -6}, {-12, 71, 66, -1}, {-14, 55, 63, 18}, {-30, 89, 72, -3}, {-5, 69, 67, -8}, {-3, 56, 78, -10}, {-12, 47, 69, 18}, 
{4, 67, 70, -12}, {-4, 68, 67, -4}, {-9, 77, 60, -2}, {-7, 65, 40, 25}, {-36, 106, 69, -11}, {-17, 74, 71, -1}, {-10, 50, 83, 2}, {-13, 61, 77, -2}, 
{11, 54, 87, -17}, {4, 58, 75, -5}, {-7, 67, 76, -6}, {-10, 75, 51, 6}, {-47, 105, 83, -7}, {-39, 90, 74, 6}, {-23, 64, 74, 13}, {-14, 48, 96, -4}, 
{-38, 108, 169, -89}, {108, -22, 126, -55}, {29, 42, 103, -28}, {4, 59, 122, -45}, {-82, 134, 101, -12}, {-45, 94, 72, 10}, {-30, 76, 65, 13}, {-17, 59, 96, -21}, 
{-18, 101, 49, -4}, {-14, 84, 64, -7}, {-25, 98, 63, -11}, {-18, 69, 73, -2}, {-17, 106, 40, -3}, {-9, 83, 50, -2}, {-11, 59, 64, 2}, {4, 46, 66, -9}, 
{-33, 117, 51, -7}, {-25, 88, 78, -16}, {-29, 92, 68, -8}, {-39, 113, 45, -4}, {-15, 98, 52, -8}, {4, 78, 55, -13}, {-18, 68, 74, -4}, {7, 27, 87, -18}, 
{-45, 121, 67, -17}, {-44, 110, 82, -23}, {-28, 92, 81, -20}, {-26, 78, 99, -27}, {-22, 101, 59, -13}, {17, 69, 63, -23}, {-1, 75, 73, -21}, {3, 43, 75, -15}, 
{-73, 148, 77, -26}, {-48, 113, 92, -32}, {-39, 105, 93, -33}, {-15, 87, 138, -65}, {-31, 111, 72, -24}, {13, 74, 67, -27}, {11, 84, 68, -30}, {-2, 88, 47, -10}, 
{7, 67, 55, -1}, {-9, 76, 60, -2}, {-9, 71, 62, 1}, {-15, 88, 37, 10}, {-36, 102, 58, 2}, {-21, 73, 61, 7}, {0, 64, 59, -8}, {-6, 47, 67, 4}, 
{-2, 58, 68, 3}, {-10, 73, 64, -2}, {-15, 68, 71, 1}, {-18, 69, 65, 4}, {-45, 103, 60, 6}, {-12, 81, 59, -8}, {-4, 74, 68, -19}, {-13, 63, 59, 7}, 
{8, 72, 72, -21}, {5, 65, 68, -8}, {-13, 80, 73, -12}, {-14, 87, 60, -10}, {-55, 135, 70, -21}, {-23, 93, 74, -18}, {-22, 74, 72, 0}, {-23, 59, 56, 25}, 
{-36, 101, 117, -32}, {103, -22, 86, -12}, {18, 48, 93, -13}, {9, 53, 88, -4}, {-82, 132, 106, -10}, {-54, 97, 94, 5}, {-37, 79, 82, 7}, {-22, 56, 74, 10}, 
{-13, 106, 37, -4}, {-31, 101, 60, -7}, {-37, 119, 61, -18}, {-27, 104, 12, 16}, {-22, 116, 36, -4}, {11, 72, 49, -10}, {6, 66, 48, -7}, {-3, 19, 79, 2}, 
{-35, 110, 53, -7}, {-35, 104, 64, -10}, {-33, 82, 89, -17}, {-47, 68, 148, -45}, {-11, 94, 47, -6}, {0, 75, 62, -13}, {5, 57, 66, -12}, {-6, 15, 96, -6}, 
{-55, 126, 58, -10}, {-74, 153, 57, -15}, {-28, 104, 65, -14}, {-79, 119, 91, -32}, {-15, 101, 44, -7}, {-7, 94, 52, -14}, {21, 64, 52, -19}, {-10, 59, 75, -12}, 
{-56, 123, 67, -16}, {-22, 111, 70, -23}, {-24, 103, 85, -31}, {0, 64, 64, 0}, {-59, 126, 58, -11}, {-27, 104, 53, -14}, {16, 67, 52, -20}, {-2, 86, 45, -12}, 
{3, 95, 33, -3}, {-15, 95, 55, -7}, {-26, 110, 54, -14}, {6, 54, 21, 33}, {-36, 127, 38, -2}, {-2, 78, 47, -4}, {-7, 52, 57, 5}, {2, 67, 27, -1}, 
{13, 72, 40, 0}, {-8, 89, 45, -4}, {-35, 109, 56, -11}, {-28, 118, 58, -25}, {-54, 132, 43, 2}, {-21, 91, 45, 2}, {-24, 97, 23, 7}, {-2, 95, 13, -10}, 
{-17, 80, 51, 6}, {-21, 93, 52, -4}, {-19, 93, 53, -7}, {-28, 100, 49, -10}, {-63, 118, 55, 8}, {-37, 96, 51, 5}, {-15, 75, 46, 1}, {-13, 83, 50, -11}, 
{-29, 93, 89, -12}, {3, 69, 74, -10}, {-4, 69, 78, -7}, {-15, 79, 91, -19}, {-84, 138, 90, -4}, {-36, 79, 91, 0}, {-14, 52, 89, -2}, {-15, 49, 91, -2}, 
{-13, 75, 75, -8}, {-10, 56, 95, -13}, {-11, 41, 112, -17}, {-10, 28, 126, -19}, {-4, 70, 65, -3}, {-4, 71, 63, -1}, {-1, 61, 63, 5}, {-3, 34, 92, 5}, 
{-21, 93, 56, 3}, {-12, 65, 81, -5}, {-12, 51, 95, -7}, {-10, 25, 114, -5}, {-2, 67, 57, 7}, {-12, 72, 66, 3}, {2, 67, 64, -6}, {0, 42, 80, 3}, 
{-25, 106, 49, 6}, {-7, 70, 76, -4}, {-7, 57, 90, -8}, {-9, 36, 86, 15}, {-6, 93, 42, 9}, {-17, 90, 57, 6}, {-22, 75, 79, 0}, {5, 47, 84, -19}, 
{-2, 73, 75, -9}, {20, 44, 111, -36}, {-3, 58, 119, -39}, {-8, 50, 111, -21}, {-63, 143, 57, 6}, {-86, 172, 129, -64}, {-32, 120, 120, -58}, {-12, 89, 99, -35}, 
{-8, 74, 76, -15}, {0, 75, 81, -27}, {-11, 77, 95, -36}, {-9, 66, 114, -47}, {-12, 75, 76, -11}, {-7, 57, 92, -14}, {-4, 45, 101, -15}, {-3, 31, 109, -10}, 
{5, 57, 94, -26}, {-9, 67, 96, -25}, {-10, 78, 93, -36}, {-7, 100, 79, -50}, {-25, 84, 93, -23}, {-17, 73, 101, -28}, {-10, 56, 115, -34}, {-6, 49, 112, -32}, 
{14, 44, 105, -26}, {17, 48, 107, -38}, {-16, 65, 121, -40}, {-8, 83, 100, -52}, {-45, 102, 110, -34}, {-26, 82, 119, -46}, {-16, 66, 128, -52}, {-10, 57, 122, -48}, 
{6, 61, 70, 2}, {113, -26, 79, -19}, {31, 42, 111, -44}, {-8, 72, 106, -41}, {-83, 144, 127, -55}, {-37, 93, 144, -70}, {-25, 77, 148, -75}, {-17, 70, 114, -50}, 
{-24, 77, 76, 0}, {-20, 76, 86, -17}, {-19, 58, 108, -24}, {-16, 41, 138, -44}, {0, 61, 71, -4}, {-5, 70, 72, -10}, {-3, 66, 73, -10}, {-7, 55, 93, -13}, 
{-25, 98, 64, -7}, {-20, 79, 79, -10}, {-17, 59, 94, -11}, {-14, 21, 140, -29}, {-7, 77, 63, -4}, {-5, 68, 68, -3}, {-1, 64, 67, -3}, {-3, 43, 87, -4}, 
{-32, 101, 46, 19}, {-17, 73, 59, 17}, {-16, 59, 78, 8}, {-10, 30, 123, -20}, {-3, 81, 46, 10}, {-4, 76, 55, 5}, {-10, 72, 68, 1}, {-2, 55, 86, -17}, 
{-7, 63, -36, 127}, {22, 24, 10, 87}, {-1, 42, 27, 71}, {-11, 39, 84, 16}, {-44, 107, -66, 156}, {-72, 141, -34, 124}, {-14, 91, -22, 100}, {-11, 75, 74, -3}, 
{-15, 94, 59, -11}, {-17, 86, 77, -21}, {-17, 94, 80, -35}, {-16, 90, 98, -55}, {-12, 88, 60, -9}, {-7, 69, 76, -10}, {-7, 64, 81, -10}, {-6, 49, 112, -33}, 
{-4, 80, 65, -12}, {-10, 80, 79, -23}, {-15, 88, 82, -32}, {-13, 114, 64, -50}, {-30, 103, 68, -12}, {-18, 82, 81, -17}, {-12, 69, 93, -24}, {-10, 60, 109, -37}, 
{-2, 74, 68, -6}, {12, 64, 73, -17}, {-13, 78, 90, -27}, {-16, 89, 102, -58}, {-46, 114, 82, -18}, {-27, 90, 92, -26}, {-16, 72, 111, -42}, {-17, 65, 140, -68}, 
{-30, 106, 42, 22}, {78, 15, 35, 16}, {22, 58, 70, -12}, {-12, 85, 65, -10}, {-79, 144, 100, -32}, {-32, 90, 115, -45}, {-32, 92, 114, -51}, {-30, 87, 113, -39}, 
{-35, 89, 81, -11}, {-34, 81, 93, -18}, {-28, 58, 115, -30}, {-21, 29, 152, -49}, {14, 52, 66, -7}, {0, 68, 70, -13}, {-1, 57, 77, -7}, {-13, 66, 98, -24}, 
{-36, 93, 78, -9}, {-29, 81, 88, -17}, {-26, 65, 101, -20}, {-19, 16, 144, -28}, {-3, 62, 75, -9}, {-5, 64, 75, -9}, {0, 58, 78, -12}, {-7, 58, 98, -30}, 
{-41, 121, 64, -14}, {-21, 84, 78, -11}, {-25, 75, 85, -9}, {-17, 40, 115, -19}, {-5, 84, 62, -11}, {-2, 77, 64, -8}, {-13, 75, 77, -12}, {-5, 51, 89, -15}, 
{-41, 104, 37, 40}, {-5, 62, 62, 18}, {-17, 63, 67, 22}, {-16, 46, 81, 17}, {-29, 103, 31, 40}, {-37, 116, 42, 26}, {-16, 99, 45, 21}, {-6, 74, 61, 5}, 
{-21, 115, 42, -11}, {-27, 106, 64, -20}, {-30, 115, 57, -26}, {-24, 106, 81, -54}, {-10, 102, 44, -10}, {-11, 84, 70, -17}, {-11, 71, 88, -22}, {-13, 62, 109, -35}, 
{-4, 91, 54, -14}, {-14, 97, 59, -17}, {-21, 104, 60, -23}, {-22, 130, 45, -46}, {-35, 114, 63, -15}, {-15, 85, 72, -14}, {-9, 77, 78, -23}, {-23, 127, 43, -24}, 
{-8, 89, 58, -8}, {11, 73, 64, -18}, {-13, 86, 78, -27}, {-14, 93, 88, -53}, {-48, 124, 69, -14}, {-22, 90, 75, -15}, {-14, 78, 84, -23}, {-14, 70, 101, -36}, 
{-28, 108, 59, -2}, {65, 33, 43, -1}, {21, 63, 69, -18}, {-9, 89, 71, -29}, {-87, 159, 83, -22}, {-37, 104, 89, -31}, {-28, 94, 93, -36}, {-39, 103, 99, -31}, 
{-51, 108, 75, -10}, {-58, 93, 97, -16}, {-55, 73, 113, -23}, {-25, -15, 174, -25}, {35, 52, 45, -6}, {17, 62, 57, -14}, {2, 66, 74, -25}, {18, 62, 37, 7}, 
{-52, 81, 98, -6}, {-47, 68, 109, -13}, {-46, 49, 126, -22}, {-19, -65, 207, -20}, {13, 22, 91, -4}, {24, 39, 67, -7}, {-3, 61, 81, -18}, {-26, 60, 125, -47}, 
{-40, 89, 84, -8}, {-45, 86, 93, -16}, {-49, 85, 94, -17}, {-22, 7, 136, -13}, {-22, 69, 81, -3}, {-7, 62, 80, -11}, {-5, 57, 82, -11}, {-10, 52, 101, -33}, 
{-38, 105, 73, -10}, {-2, 57, 87, -15}, {-23, 71, 91, -16}, {-19, 52, 101, -15}, {-77, 145, 68, -2}, {-59, 133, 63, 3}, {-7, 92, 53, 8}, {-18, 92, 78, -15}, 
{-23, 129, 29, -10}, {-44, 140, 39, -16}, {-50, 139, 43, -20}, {-24, 155, 4, -25}, {13, 94, 29, -10}, {6, 74, 59, -15}, {-2, 103, 41, -23}, {13, 18, 99, -18}, 
{-4, 118, 20, -9}, {-24, 130, 28, -14}, {-31, 147, 15, -18}, {-18, 201, -57, -20}, {-11, 99, 51, -12}, {-26, 100, 71, -18}, {4, 32, 120, -44}, {-43, 168, 44, -53}, 
{9, 92, 35, -9}, {-19, 123, 28, -10}, {-13, 110, 40, -17}, {-13, 137, 6, -23}, {-50, 131, 64, -15}, {-44, 112, 83, -22}, {-18, 89, 78, -30}, {-8, 75, 79, -45}, 
{-11, 103, 49, -8}, {22, 79, 40, -10}, {11, 83, 51, -16}, {-16, 102, 53, -21}, {-74, 149, 78, -20}, {-42, 120, 77, -35}, {-63, 146, 76, -33}, {-51, 155, 16, 34} };

static int tapsD2[ 512 ][ 4 ] = { {-14, 78, 77, -13}, {-8, 61, 87, -12}, {-7, 48, 92, -6}, {-9, 39, 89, 7}, {-12, 81, 73, -14}, {-2, 72, 92, -34}, {5, 57, 107, -43}, {-2, 39, 117, -28}, 
{-23, 92, 84, -23}, {-10, 68, 95, -24}, {-6, 60, 96, -20}, {-12, 59, 92, -12}, {-6, 76, 89, -29}, {-14, 73, 102, -32}, {7, 58, 104, -45}, {-1, 22, 150, -51}, 
{-28, 102, 94, -35}, {-16, 79, 104, -36}, {-9, 64, 113, -37}, {-16, 69, 100, -26}, {-6, 83, 99, -42}, {-14, 84, 105, -41}, {-19, 67, 124, -43}, {1, 38, 115, -40}, 
{-19, 82, 150, -81}, {-10, 74, 133, -62}, {-11, 67, 142, -66}, {-20, 72, 126, -50}, {-57, 128, 161, -95}, {-46, 133, 146, -86}, {-12, 109, 106, -55}, {-12, 85, 136, -70}, 
{-7, 71, 71, -7}, {-9, 73, 59, 3}, {-8, 55, 58, 20}, {-6, 41, 59, 32}, {-15, 75, 79, -10}, {-3, 54, 98, -22}, {0, 39, 110, -23}, {-3, 40, 107, -17}, 
{-3, 68, 71, -8}, {-7, 73, 72, -11}, {-6, 83, 44, 2}, {-7, 82, 30, 18}, {-28, 89, 76, -7}, {-16, 72, 81, -9}, {-4, 49, 101, -20}, {-4, 46, 99, -17}, 
{9, 48, 86, -8}, {2, 59, 71, -1}, {-10, 69, 72, -4}, {-9, 85, 54, -7}, {-41, 97, 82, -4}, {-28, 85, 80, -8}, {-12, 62, 100, -27}, {-2, 42, 122, -38}, 
{37, 25, 155, -71}, {125, -45, 94, -22}, {20, 47, 116, -43}, {-12, 71, 148, -72}, {-95, 157, 142, -64}, {-60, 127, 93, -30}, {-28, 89, 114, -51}, {-7, 56, 133, -66}, 
{-13, 88, 58, -6}, {-13, 77, 79, -14}, {-12, 62, 88, -11}, {-23, 59, 73, 10}, {-21, 99, 52, -4}, {-13, 73, 71, -7}, {6, 59, 79, -24}, {-5, 36, 85, 1}, 
{-26, 93, 79, -18}, {-21, 87, 96, -31}, {-3, 62, 96, -23}, {-18, 72, 93, -20}, {-12, 84, 71, -16}, {-7, 67, 88, -22}, {-9, 56, 96, -22}, {-3, 32, 96, -12}, 
{-38, 106, 86, -25}, {-25, 91, 90, -26}, {-14, 78, 94, -27}, {-19, 82, 103, -36}, {-11, 82, 86, -28}, {11, 62, 99, -42}, {-20, 68, 115, -34}, {-1, 44, 102, -30}, 
{-46, 109, 115, -49}, {-35, 96, 115, -47}, {-26, 82, 112, -41}, {-28, 86, 100, -32}, {-39, 100, 125, -57}, {-22, 95, 128, -65}, {1, 91, 126, -73}, {-12, 79, 100, -36}, 
{0, 65, 65, -3}, {-13, 75, 67, -3}, {-12, 67, 70, 1}, {-18, 85, 39, 13}, {-35, 93, 71, -1}, {-9, 73, 69, -10}, {-1, 54, 82, -13}, {-7, 63, 73, -6}, 
{-1, 68, 70, -8}, {-12, 79, 69, -9}, {-11, 82, 54, -1}, {-12, 85, 21, 24}, {-32, 101, 72, -13}, {-21, 88, 67, -9}, {-10, 61, 73, -1}, {-9, 56, 81, -6}, 
{6, 66, 83, -23}, {-5, 66, 74, -5}, {-6, 68, 64, 2}, {-15, 85, 47, 2}, {-41, 106, 84, -17}, {-41, 100, 63, 9}, {-27, 73, 56, 22}, {-7, 41, 102, -13}, 
{-36, 95, 163, -69}, {117, -39, 117, -38}, {8, 45, 105, -16}, {-24, 72, 149, -60}, {-64, 120, 146, -52}, {-63, 122, 95, -15}, {-44, 102, 72, -6}, {-12, 66, 108, -52}, 
{-12, 93, 45, -1}, {-18, 80, 71, -8}, {-14, 75, 64, -1}, {-17, 61, 89, -9}, {-23, 111, 38, 0}, {-14, 84, 53, -1}, {-16, 64, 62, 5}, {1, 56, 55, -6}, 
{-24, 99, 54, -5}, {-26, 101, 61, -10}, {-18, 89, 69, -13}, {-8, 122, 50, -24}, {-24, 105, 48, -4}, {0, 75, 58, -10}, {-21, 63, 79, -3}, {-2, 12, 112, -20}, 
{-41, 111, 67, -13}, {-29, 93, 88, -25}, {-23, 85, 108, -39}, {-21, 84, 91, -25}, {-30, 104, 62, -11}, {21, 58, 72, -26}, {-1, 68, 83, -25}, {-10, 39, 102, -21}, 
{-53, 118, 89, -28}, {-42, 104, 94, -31}, {-38, 98, 91, -28}, {-21, 76, 104, -32}, {-51, 112, 92, -30}, {-13, 74, 106, -44}, {11, 65, 117, -61}, {-16, 79, 76, -15}, 
{7, 72, 47, 0}, {-11, 76, 61, -1}, {-19, 86, 52, 1}, {-40, 41, 124, -11}, {-41, 109, 55, 3}, {-25, 81, 58, 7}, {2, 64, 61, -13}, {-9, 65, 54, 3}, 
{4, 55, 58, 9}, {-9, 71, 59, 3}, {-12, 87, 42, 2}, {-27, 142, 18, -21}, {-47, 103, 58, 9}, {-25, 94, 55, -3}, {-2, 82, 60, -22}, {-16, 89, 41, 1}, 
{0, 77, 58, -8}, {-6, 77, 61, -7}, {-19, 86, 64, -9}, {-24, 96, 48, -5}, {-43, 123, 65, -17}, {-30, 112, 67, -22}, {-24, 84, 70, -6}, {-20, 58, 54, 22}, 
{-43, 96, 113, -16}, {93, -29, 76, 12}, {20, 26, 86, 14}, {-18, 61, 98, -2}, {-56, 107, 107, -11}, {-64, 117, 93, 0}, {-56, 109, 68, 12}, {-25, 68, 52, 15}, 
{-5, 93, 38, -2}, {-47, 126, 50, -9}, {-28, 78, 77, -8}, {-20, 50, 88, -2}, {-16, 105, 41, -3}, {-4, 73, 62, -8}, {3, 59, 61, -10}, {-4, 38, 71, -5}, 
{-23, 88, 63, -6}, {-46, 110, 70, -12}, {-10, 14, 141, -22}, {-89, -4, 211, -39}, {-18, 99, 46, -4}, {-14, 89, 58, -9}, {2, 46, 85, -16}, {-12, 24, 100, -8}, 
{-24, 97, 60, -9}, {-43, 107, 75, -17}, {-29, 102, 64, -13}, {-73, 117, 93, -35}, {-40, 124, 43, -3}, {-21, 109, 40, -6}, {25, 52, 59, -22}, {-14, 62, 79, -13}, 
{-49, 119, 66, -14}, {-55, 115, 83, -24}, {-42, 107, 90, -29}, {100, -22, 78, 13}, {-63, 131, 58, -8}, {-51, 105, 68, -13}, {10, 54, 70, -25}, {-3, 65, 75, -23}, 
{3, 95, 31, -2}, {-9, 102, 39, -6}, {-19, 103, 51, -10}, {-5, 139, 7, -14}, {-26, 114, 41, -2}, {1, 85, 40, -6}, {-5, 58, 55, -1}, {-4, 70, 38, -4}, 
{-4, 91, 35, 1}, {-6, 92, 34, 0}, {-34, 98, 51, -2}, {-59, 86, 92, -14}, {-38, 133, 28, -2}, {-9, 92, 33, -3}, {-20, 110, 14, -1}, {-9, 92, 28, -8}, 
{-49, 119, 38, 5}, {-28, 97, 42, 2}, {-34, 97, 48, -2}, {-28, 116, 32, -13}, {-47, 123, 36, 2}, {-27, 101, 40, -1}, {-25, 103, 35, -5}, {-12, 75, 64, -12}, 
{-27, 98, 84, -16}, {-3, 74, 70, -10}, {-10, 78, 76, -13}, {-17, 83, 98, -26}, {-68, 130, 84, -9}, {-27, 90, 79, -14}, {-13, 72, 76, -13}, {-21, 74, 64, -4}, 
{-16, 77, 81, -13}, {-12, 59, 103, -24}, {-10, 44, 114, -23}, {-10, 31, 108, -4}, {-7, 72, 70, -7}, {-3, 67, 64, -1}, {0, 47, 71, 8}, {-2, 32, 95, 1}, 
{-24, 92, 66, -5}, {-13, 67, 86, -12}, {-7, 45, 103, -14}, {-8, 21, 103, 8}, {-8, 74, 65, -2}, {-10, 72, 69, -2}, {8, 61, 58, -1}, {2, 35, 88, -2}, 
{-17, 96, 55, 2}, {-9, 75, 76, -9}, {-3, 54, 97, -17}, {-7, 38, 102, -5}, {-12, 91, 54, 2}, {-24, 88, 69, 1}, {-10, 61, 79, -1}, {5, 40, 117, -47}, 
{12, 56, 41, 29}, {49, 9, 102, -21}, {7, 49, 119, -38}, {-9, 55, 121, -37}, {-67, 149, 36, 26}, {-74, 165, 88, -27}, {-21, 115, 90, -33}, {-15, 86, 98, -28}, 
{-12, 79, 77, -17}, {-8, 82, 87, -34}, {-11, 77, 102, -44}, {-9, 69, 98, -35}, {-12, 76, 77, -13}, {-5, 57, 88, -13}, {-1, 45, 93, -11}, {-2, 39, 93, -5}, 
{-4, 66, 92, -24}, {-11, 72, 92, -24}, {-9, 84, 87, -38}, {-9, 99, 80, -50}, {-23, 83, 92, -23}, {-17, 76, 94, -25}, {-3, 49, 103, -24}, {-7, 66, 85, -22}, 
{7, 50, 97, -18}, {3, 58, 96, -25}, {-14, 71, 105, -34}, {-13, 88, 97, -51}, {-33, 91, 103, -28}, {-27, 87, 104, -36}, {-13, 66, 110, -39}, {-7, 58, 101, -29}, 
{49, 20, 68, 3}, {107, -23, 71, -11}, {19, 51, 94, -29}, {-16, 75, 120, -50}, {-78, 146, 81, -16}, {-50, 116, 109, -45}, {-29, 90, 117, -53}, {-14, 64, 122, -50}, 
{-32, 85, 80, -7}, {-22, 78, 92, -23}, {-18, 65, 110, -33}, {-13, 44, 153, -65}, {0, 58, 77, -8}, {-4, 69, 75, -13}, {0, 55, 79, -10}, {-5, 36, 100, -4}, 
{-26, 93, 74, -13}, {-24, 84, 82, -15}, {-15, 53, 103, -17}, {-12, 26, 151, -47}, {-8, 73, 73, -9}, {-6, 68, 75, -10}, {1, 60, 76, -14}, {-1, 32, 100, -12}, 
{-29, 100, 55, 7}, {-21, 84, 59, 10}, {-19, 65, 74, 8}, {-12, 32, 121, -21}, {-3, 72, 58, 4}, {-3, 70, 64, -1}, {-3, 57, 69, 3}, {2, 37, 95, -22}, 
{-6, 70, -12, 92}, {8, 48, 15, 72}, {-9, 53, 21, 73}, {-11, 49, 84, 8}, {-31, 106, -32, 108}, {-37, 119, -25, 100}, {11, 75, -43, 112}, {-12, 74, 71, 1}, 
{-25, 102, 62, -13}, {-23, 90, 80, -23}, {-17, 86, 92, -39}, {-13, 87, 116, -73}, {-10, 86, 60, -8}, {-13, 77, 78, -14}, {-5, 62, 84, -16}, {-9, 50, 124, -44}, 
{-10, 84, 69, -14}, {-16, 81, 88, -26}, {-15, 86, 85, -33}, {-13, 94, 101, -66}, {-24, 95, 69, -10}, {-29, 93, 85, -19}, {-12, 63, 102, -27}, {-9, 62, 106, -38}, 
{-4, 71, 77, -10}, {7, 63, 82, -19}, {-19, 74, 102, -29}, {-14, 86, 101, -53}, {-36, 103, 80, -17}, {-28, 91, 91, -26}, {-26, 89, 92, -28}, {-15, 71, 109, -44}, 
{-3, 86, 9, 49}, {80, 9, 41, 12}, {13, 64, 61, -3}, {-28, 92, 67, -6}, {-62, 132, 69, -6}, {-51, 120, 95, -35}, {-31, 94, 103, -40}, {-19, 74, 114, -49}, 
{-45, 99, 81, -11}, {-39, 89, 88, -15}, {-27, 69, 110, -35}, {-27, 52, 134, -49}, {19, 57, 58, -8}, {3, 68, 65, -11}, {2, 66, 72, -20}, {-8, 23, 126, -18}, 
{-35, 86, 84, -10}, {-30, 82, 88, -16}, {-23, 60, 110, -28}, {-23, 24, 143, -34}, {-4, 49, 85, -6}, {2, 53, 79, -10}, {3, 56, 78, -16}, {-6, 38, 113, -31}, 
{-32, 104, 72, -15}, {-30, 101, 74, -18}, {-25, 75, 90, -17}, {-19, 35, 123, -24}, {-8, 79, 70, -12}, {3, 65, 69, -10}, {-12, 60, 90, -15}, {-5, 56, 102, -40}, 
{-32, 100, 50, 17}, {6, 52, 68, 9}, {-25, 78, 69, 7}, {-21, 61, 89, -5}, {-48, 122, 46, 20}, {-13, 97, 48, 7}, {16, 79, 28, 21}, {-12, 82, 85, -22}, 
{-27, 117, 46, -10}, {-34, 107, 69, -19}, {-36, 109, 72, -28}, {-23, 118, 73, -58}, {-5, 92, 47, -7}, {-11, 86, 65, -13}, {-5, 72, 75, -18}, {-11, 97, 63, -29}, 
{-14, 104, 44, -7}, {-18, 101, 58, -15}, {-30, 108, 63, -24}, {-19, 114, 71, -55}, {-19, 96, 59, -6}, {-22, 93, 64, -3}, {-15, 75, 86, -20}, {-18, 101, 63, -23}, 
{-9, 91, 54, -4}, {4, 78, 66, -19}, {-18, 90, 77, -27}, {-26, 99, 80, -39}, {-37, 112, 65, -9}, {-29, 97, 76, -13}, {-41, 111, 84, -22}, {-8, 56, 96, -19}, 
{-37, 118, 47, 9}, {78, 16, 56, -12}, {7, 71, 77, -25}, {-31, 93, 86, -28}, {-66, 142, 66, -9}, {-43, 114, 81, -25}, {-31, 94, 100, -37}, {-25, 81, 114, -44}, 
{-41, 100, 74, -10}, {-72, 117, 85, -14}, {-55, 72, 117, -25}, {-28, 4, 165, -34}, {30, 62, 40, -6}, {24, 31, 79, -15}, {-5, 97, 59, -37}, {3, -1, 118, 5}, 
{-53, 85, 97, -10}, {-63, 96, 96, -14}, {-55, 71, 114, -20}, {-18, -45, 195, -27}, {15, 26, 88, -7}, {18, 22, 89, -11}, {-10, 15, 131, -24}, {2, 102, 43, -32}, 
{-58, 108, 85, -14}, {-59, 110, 83, -14}, {-41, 82, 98, -25}, {-16, 13, 137, -26}, {-2, 53, 83, -13}, {19, 32, 81, -11}, {-11, 52, 100, -25}, {-11, 28, 110, -16}, 
{-40, 106, 75, -11}, {-6, 66, 89, -23}, {-32, 84, 90, -22}, {-12, 59, 97, -19}, {-68, 140, 70, -8}, {-46, 137, 66, -18}, {10, 89, 58, -18}, {-25, 94, 87, -19}, 
{-10, 111, 33, -11}, {-68, 155, 44, -13}, {-51, 139, 49, -28}, {-36, 166, 4, -27}, {8, 89, 38, -10}, {15, 79, 46, -21}, {-8, 67, 82, -15}, {17, 27, 81, -6}, 
{6, 105, 22, -8}, {-47, 154, 23, -13}, {-35, 144, 25, -24}, {-30, 197, -45, -18}, {-8, 93, 53, -11}, {-8, 80, 68, -12}, {-34, 97, 84, -13}, {-13, 82, 93, -64}, 
{-1, 99, 35, -6}, {-21, 121, 32, -12}, {-25, 122, 37, -20}, {-26, 139, 11, -15}, {-27, 104, 66, -15}, {-36, 105, 80, -20}, {-33, 109, 67, -9}, {-51, 106, 146, -103}, 
{-40, 124, 54, -7}, {13, 80, 48, -12}, {-7, 91, 61, -21}, {-19, 97, 59, -12}, {-67, 146, 68, -19}, {-40, 108, 87, -25}, {-34, 98, 89, -14}, {15, 75, -18, 87} };

static int tapsC1[ 64 ][ 2 ] = { {60, 68}, {48, 81}, {34, 95}, {18, 111}, {61, 67}, {52, 76}, {45, 84}, {24, 105}, 
{68, 60}, {56, 72}, {32, 96}, {14, 114}, {70, 58}, {62, 66}, {48, 80}, {34, 95}, 
{78, 51}, {76, 53}, {48, 80}, {10, 118}, {77, 52}, {77, 51}, {65, 64}, {43, 86}, 
{86, 43}, {81, 47}, {81, 46}, {40, 87}, {87, 42}, {87, 42}, {82, 48}, {59, 71}, 
{60, 68}, {52, 76}, {42, 87}, {24, 106}, {59, 69}, {48, 80}, {36, 92}, {17, 111}, 
{70, 59}, {63, 65}, {46, 83}, {35, 94}, {68, 60}, {58, 70}, {37, 91}, {14, 113}, 
{80, 48}, {76, 53}, {65, 64}, {40, 89}, {78, 50}, {76, 53}, {49, 79}, {13, 115}, 
{84, 46}, {88, 42}, {87, 43}, {61, 69}, {89, 40}, {87, 41}, {89, 39}, {41, 87} };

static int tapsC2[ 64 ][ 2 ] = { {4, 124}, {-1, 129}, {-1, 130}, {1, 127}, {4, 124}, {-2, 130}, {1, 128}, {0, 128}, 
{-4, 132}, {-1, 129}, {0, 128}, {2, 125}, {-5, 133}, {-3, 131}, {-4, 132}, {1, 126}, 
{-2, 131}, {2, 127}, {2, 126}, {5, 123}, {-7, 135}, {-5, 133}, {-7, 135}, {5, 121}, 
{-3, 134}, {4, 125}, {3, 126}, {4, 124}, {-6, 136}, {-2, 132}, {-5, 134}, {2, 126}, 
{5, 123}, {-1, 129}, {-1, 130}, {-3, 131}, {5, 123}, {0, 129}, {-1, 129}, {3, 125}, 
{-2, 130}, {-1, 129}, {0, 128}, {5, 122}, {-2, 130}, {-3, 130}, {0, 128}, {1, 127}, 
{-4, 132}, {-1, 130}, {-3, 131}, {11, 116}, {-1, 130}, {3, 125}, {-1, 128}, {-3, 131}, 
{0, 130}, {8, 121}, {2, 127}, {5, 122}, {0, 130}, {11, 118}, {8, 120}, {2, 126} };

static const int TAP_SHIFT = 7;

static const int OFFSET = (1<<(TAP_SHIFT-1));

int FilterRound(const int* vals, const int* taps, const int size){

  int ret_val = OFFSET;
  for (int j=0; j<size; ++j){
    ret_val += vals[j]*taps[j];
  }
  return (ret_val >> TAP_SHIFT);
}

void upScaleComponent2x2Adaptive( Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){

  // Simplified version of: Shi et al,"Context-based adaptive image resolution upconversion", Journal of Electronic Imaging 19(1), 013008 (Jan–Mar 2010)

  const int iDestHeight = iSrcHeight*2;
  const int iDestWidth = iSrcWidth*2;

  int cand_vec[4];
  int cand_vec_mix[2];

  int idx1, idx2,idxc;
  int d[3];
  int val1=0;
  int val2=0;

  const Pel* pDataSrc = pSrc - 4*iSrcStride - 4;
  Pel* pDataDest= pDest - 8*iDestStride - 8;

  /* Position of sample are:
    C H C H C H C H C H
    V D V D V D V D V D
    C H C H C H C H C H
    V D V D V D V D V D
   C=Co-sited, H=horizontal, V=vertical, D=diagonal
  */


  // Begin with copy into C postions
  for (int j=0, jj=0; j<iSrcHeight+8; ++j, jj+=2){
    for (int i=0,ii=0; i<iSrcWidth+8; ++i, ii+=2){
      pDataDest[ii] = pDataSrc[i];
    }
    pDataDest += 2*iDestStride;
    pDataSrc += iSrcStride;
  }

  pDataDest = pDest-3*iDestStride-4;

  // Next, do D positions
  for (int j=-3; j<iDestHeight+3; j+=2){
    for (int i=1; i<iDestWidth+8; i+=2){

      d[0] = (pDataDest[i-iDestStride-1] - pDataDest[i-3*iDestStride-3])>>g_uiBitIncrement;
      d[1] = (pDataDest[i+iDestStride+1] - pDataDest[i-iDestStride-1])>>g_uiBitIncrement;
      d[2] = (pDataDest[i+3*iDestStride+3] - pDataDest[i+iDestStride+1])>>g_uiBitIncrement;
      idx1 = getClassIndex(d,3);

      cand_vec[0] = pDataDest[i-3*iDestStride-3];
      cand_vec[1] = pDataDest[i-iDestStride-1];
      cand_vec[2] = pDataDest[i+iDestStride+1];
      cand_vec[3] = pDataDest[i+3*iDestStride+3];
      val1 = Clip( FilterRound(cand_vec, tapsD1[idx1], 4) );

      d[0] = (pDataDest[i-iDestStride+1] - pDataDest[i-3*iDestStride+3])>>g_uiBitIncrement;
      d[1] = (pDataDest[i+iDestStride-1] - pDataDest[i-iDestStride+1])>>g_uiBitIncrement;
      d[2] = (pDataDest[i+3*iDestStride-3] - pDataDest[i+iDestStride-1])>>g_uiBitIncrement;
      idx2 = getClassIndex(d,3);

      cand_vec[0] = pDataDest[i-3*iDestStride+3];
      cand_vec[1] = pDataDest[i-iDestStride+1];
      cand_vec[2] = pDataDest[i+iDestStride-1];
      cand_vec[3] = pDataDest[i+3*iDestStride-3];
      val2 = Clip( FilterRound(cand_vec, tapsD1[idx2], 4) );

      d[0] = (pDataDest[i-iDestStride-1] - pDataDest[i+iDestStride+1])>>g_uiBitIncrement;
      d[1] = (pDataDest[i-iDestStride+1] - pDataDest[i+iDestStride-1])>>g_uiBitIncrement;
      idxc = getClassIndex(d,2);

      cand_vec_mix[0] = val1;
      cand_vec_mix[1] = val2;
      pDataDest[i] = Clip( FilterRound( cand_vec_mix, tapsC1[idxc], 2) );
    }
    pDataDest += 2*iDestStride;
  }

  pDataDest = pDest;

  // Next, do V and H points
  for (int j=0; j<iDestHeight; j++){
    for (int i=((j+1)%2); i<iDestWidth; i+=2){

      d[0] = (pDataDest[i-iDestStride] - pDataDest[i-3*iDestStride])>>g_uiBitIncrement;
      d[1] = (pDataDest[i+iDestStride] - pDataDest[i-iDestStride])>>g_uiBitIncrement;
      d[2] = (pDataDest[i+3*iDestStride] - pDataDest[i+iDestStride])>>g_uiBitIncrement;
      idx1 = getClassIndex(d,3);

      cand_vec[0] = pDataDest[i-3*iDestStride];
      cand_vec[1] = pDataDest[i-iDestStride];
      cand_vec[2] = pDataDest[i+iDestStride];
      cand_vec[3] = pDataDest[i+3*iDestStride];
      val1 = Clip( FilterRound(cand_vec, tapsD2[idx1], 4) );

      d[0] = (pDataDest[i-1] - pDataDest[i-3])>>g_uiBitIncrement;
      d[1] = (pDataDest[i+1] - pDataDest[i-1])>>g_uiBitIncrement;
      d[2] = (pDataDest[i+3] - pDataDest[i+1])>>g_uiBitIncrement;
      idx2 = getClassIndex(d,3);

      cand_vec[0] = pDataDest[i-3];
      cand_vec[1] = pDataDest[i-1];
      cand_vec[2] = pDataDest[i+1];
      cand_vec[3] = pDataDest[i+3];
      val2 = Clip( FilterRound( cand_vec, tapsD2[idx2], 4) );

      d[0] = (pDataDest[i-iDestStride] - pDataDest[i+iDestStride])>>g_uiBitIncrement;
      d[1] = (pDataDest[i+1] - pDataDest[i-1])>>g_uiBitIncrement;
      idxc = getClassIndex(d,2);

      if (j%2==0){
       cand_vec_mix[0] = val1;
       cand_vec_mix[1] = val2;
      } else {
       cand_vec_mix[0] = val2;
       cand_vec_mix[1] = val1;
      }

      pDataDest[i] = Clip( FilterRound(cand_vec_mix, tapsC2[idxc], 2) );
    }
    pDataDest += iDestStride;
  }

}



void upScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){
  const int iDestHeight = iSrcHeight*2;

  const int NUM_UP_TAPS = 2;
  const int iTaps[NUM_UP_TAPS] = {20, -4};
  const int iFilterShift = 5;

  const int iTempStride = iSrcWidth+2*NUM_UP_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];
  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + NUM_UP_TAPS;
  int iTemp;

  for (int j=0; j<iSrcHeight; ++j){
    // Zero phase
    for (int i=0; i<iSrcWidth; ++i){
      pDataDest[i] = pDataSrc[i];
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );
    pDataDest += iTempStride;

    // +1/2 phase
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i + (k+1)*iSrcStride]  + pDataSrc[i - k*iSrcStride] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );

    pDataDest += iTempStride;
    pDataSrc += iSrcStride;
  }

  pDataSrc = pTemp + NUM_UP_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iSrcWidth; ++i, ii+=2){
      pDataDest[ii] = pDataSrc[i];
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i - k] + pDataSrc[i + k+1] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[ii+1] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

void upScaleAdaptive(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==1 && iDestResIdx == 0){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    upScaleComponent2x2Adaptive( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
    upScaleComponent2x2Adaptive( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    upScaleComponent2x2Adaptive( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}


void upScale(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==1 && iDestResIdx == 0){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    upScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
    upScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    upScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}


void downScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){

  const int iDestWidth = iSrcWidth/2;
  const int iDestHeight = iSrcHeight/2;

  const int HALF_NUM_DOWN_TAPS = 5;
  const int iTaps[HALF_NUM_DOWN_TAPS] = {40, 0, -12, 0, 4};
  const int iCentreTap = 64;
  const int iFilterShift = 7;

  const int iTempStride = iSrcWidth + 2*HALF_NUM_DOWN_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];

  int iTemp;

  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + HALF_NUM_DOWN_TAPS;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = pDataSrc[i]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[i+(k+1)*iSrcStride] + pDataSrc[i-(k+1)*iSrcStride]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - HALF_NUM_DOWN_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+HALF_NUM_DOWN_TAPS, pDataDest[iSrcWidth-1] );
    pDataSrc += 2*iSrcStride;
    pDataDest += iTempStride;
  }

  // Horizontal
  pDataSrc = pTemp+HALF_NUM_DOWN_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iDestWidth; ++i,ii+=2){
      iTemp = 0;
      iTemp = pDataSrc[ii]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[ii+k+1] + pDataSrc[ii-k-1]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

void downScale(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==0 && iDestResIdx == 1){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    downScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
    downScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    downScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}

